@@ -27,14 +27,18 @@
 #include <memory.h>
 #include <math.h>
 #include <glib/gi18n.h>
+#include <libxklavier/xklavier.h>
 
 #include <gkbd-keyboard-drawing.h>
 #include <gkbd-keyboard-drawing-marshal.h>
+#include <gkbd-util.h>
 
 #define noKBDRAW_DEBUG
 
 #define INVALID_KEYCODE ((guint)(-1))
 
+#define GTK_RESPONSE_PRINT 2
+
 enum {
 	BAD_KEYCODE = 0,
 	NUM_SIGNALS
@@ -616,147 +620,147 @@ set_key_label_in_layout (GkbdKeyboardDrawingRenderContext * context,
 	PangoLayout *layout = context->layout;
 
 	switch (keyval) {
-	case GDK_Scroll_Lock:
+	case GDK_KEY_Scroll_Lock:
 		pango_layout_set_text (layout, "Scroll\nLock", -1);
 		break;
 
-	case GDK_space:
+	case GDK_KEY_space:
 		pango_layout_set_text (layout, "", -1);
 		break;
 
-	case GDK_Sys_Req:
+	case GDK_KEY_Sys_Req:
 		pango_layout_set_text (layout, "Sys Rq", -1);
 		break;
 
-	case GDK_Page_Up:
+	case GDK_KEY_Page_Up:
 		pango_layout_set_text (layout, "Page\nUp", -1);
 		break;
 
-	case GDK_Page_Down:
+	case GDK_KEY_Page_Down:
 		pango_layout_set_text (layout, "Page\nDown", -1);
 		break;
 
-	case GDK_Num_Lock:
+	case GDK_KEY_Num_Lock:
 		pango_layout_set_text (layout, "Num\nLock", -1);
 		break;
 
-	case GDK_KP_Page_Up:
+	case GDK_KEY_KP_Page_Up:
 		pango_layout_set_text (layout, "Pg Up", -1);
 		break;
 
-	case GDK_KP_Page_Down:
+	case GDK_KEY_KP_Page_Down:
 		pango_layout_set_text (layout, "Pg Dn", -1);
 		break;
 
-	case GDK_KP_Home:
+	case GDK_KEY_KP_Home:
 		pango_layout_set_text (layout, "Home", -1);
 		break;
 
-	case GDK_KP_Left:
+	case GDK_KEY_KP_Left:
 		pango_layout_set_text (layout, "Left", -1);
 		break;
 
-	case GDK_KP_End:
+	case GDK_KEY_KP_End:
 		pango_layout_set_text (layout, "End", -1);
 		break;
 
-	case GDK_KP_Up:
+	case GDK_KEY_KP_Up:
 		pango_layout_set_text (layout, "Up", -1);
 		break;
 
-	case GDK_KP_Begin:
+	case GDK_KEY_KP_Begin:
 		pango_layout_set_text (layout, "Begin", -1);
 		break;
 
-	case GDK_KP_Right:
+	case GDK_KEY_KP_Right:
 		pango_layout_set_text (layout, "Right", -1);
 		break;
 
-	case GDK_KP_Enter:
+	case GDK_KEY_KP_Enter:
 		pango_layout_set_text (layout, "Enter", -1);
 		break;
 
-	case GDK_KP_Down:
+	case GDK_KEY_KP_Down:
 		pango_layout_set_text (layout, "Down", -1);
 		break;
 
-	case GDK_KP_Insert:
+	case GDK_KEY_KP_Insert:
 		pango_layout_set_text (layout, "Ins", -1);
 		break;
 
-	case GDK_KP_Delete:
+	case GDK_KEY_KP_Delete:
 		pango_layout_set_text (layout, "Del", -1);
 		break;
 
-	case GDK_dead_grave:
+	case GDK_KEY_dead_grave:
 		pango_layout_set_text (layout, "ˋ", -1);
 		break;
 
-	case GDK_dead_acute:
+	case GDK_KEY_dead_acute:
 		pango_layout_set_text (layout, "ˊ", -1);
 		break;
 
-	case GDK_dead_circumflex:
+	case GDK_KEY_dead_circumflex:
 		pango_layout_set_text (layout, "ˆ", -1);
 		break;
 
-	case GDK_dead_tilde:
+	case GDK_KEY_dead_tilde:
 		pango_layout_set_text (layout, "~", -1);
 		break;
 
-	case GDK_dead_macron:
+	case GDK_KEY_dead_macron:
 		pango_layout_set_text (layout, "ˉ", -1);
 		break;
 
-	case GDK_dead_breve:
+	case GDK_KEY_dead_breve:
 		pango_layout_set_text (layout, "˘", -1);
 		break;
 
-	case GDK_dead_abovedot:
+	case GDK_KEY_dead_abovedot:
 		pango_layout_set_text (layout, "˙", -1);
 		break;
 
-	case GDK_dead_diaeresis:
+	case GDK_KEY_dead_diaeresis:
 		pango_layout_set_text (layout, "¨", -1);
 		break;
 
-	case GDK_dead_abovering:
+	case GDK_KEY_dead_abovering:
 		pango_layout_set_text (layout, "˚", -1);
 		break;
 
-	case GDK_dead_doubleacute:
+	case GDK_KEY_dead_doubleacute:
 		pango_layout_set_text (layout, "˝", -1);
 		break;
 
-	case GDK_dead_caron:
+	case GDK_KEY_dead_caron:
 		pango_layout_set_text (layout, "ˇ", -1);
 		break;
 
-	case GDK_dead_cedilla:
+	case GDK_KEY_dead_cedilla:
 		pango_layout_set_text (layout, "¸", -1);
 		break;
 
-	case GDK_dead_ogonek:
+	case GDK_KEY_dead_ogonek:
 		pango_layout_set_text (layout, "˛", -1);
 		break;
 
-		/* case GDK_dead_iota:
-		 * case GDK_dead_voiced_sound:
-		 * case GDK_dead_semivoiced_sound: */
+		/* case GDK_KEY_dead_iota:
+		 * case GDK_KEY_dead_voiced_sound:
+		 * case GDK_KEY_dead_semivoiced_sound: */
 
-	case GDK_dead_belowdot:
+	case GDK_KEY_dead_belowdot:
 		pango_layout_set_text (layout, " ̣", -1);
 		break;
 
-	case GDK_horizconnector:
+	case GDK_KEY_horizconnector:
 		pango_layout_set_text (layout, "horiz\nconn", -1);
 		break;
 
-	case GDK_Mode_switch:
+	case GDK_KEY_Mode_switch:
 		pango_layout_set_text (layout, "AltGr", -1);
 		break;
 
-	case GDK_Multi_key:
+	case GDK_KEY_Multi_key:
 		pango_layout_set_text (layout, "Compose", -1);
 		break;
 
@@ -998,8 +1002,8 @@ draw_key (GkbdKeyboardDrawingRenderContext * context,
 
 	if (key->pressed)
 		color =
-		    &(GTK_WIDGET (drawing)->
-		      style->base[GTK_STATE_SELECTED]);
+		    &gtk_widget_get_style (GTK_WIDGET (drawing))->base
+		    [GTK_STATE_SELECTED];
 	else
 		color = drawing->colors + key->xkbkey->color_ndx;
 
@@ -1097,8 +1101,9 @@ invalidate_indicator_doodad_region (GkbdKeyboardDrawing * drawing,
 			   doodad->doodad->indicator.left,
 			   doodad->origin_y +
 			   doodad->doodad->indicator.top,
-			   &drawing->xkb->geom->shapes[doodad->
-						       doodad->indicator.shape_ndx]);
+			   &drawing->xkb->geom->shapes[doodad->doodad->
+						       indicator.
+						       shape_ndx]);
 }
 
 static void
@@ -1112,8 +1117,8 @@ invalidate_key_region (GkbdKeyboardDrawing * drawing,
 			   key->angle,
 			   key->origin_x,
 			   key->origin_y,
-			   &drawing->xkb->geom->shapes[key->
-						       xkbkey->shape_ndx]);
+			   &drawing->xkb->geom->shapes[key->xkbkey->
+						       shape_ndx]);
 }
 
 static void
@@ -1293,14 +1298,19 @@ draw_keyboard_to_context (GkbdKeyboardDrawingRenderContext * context,
 			(GFunc) draw_keyboard_item, &data);
 }
 
-static void
+static gboolean
 create_cairo (GkbdKeyboardDrawing * drawing)
 {
-	GtkStateType state = GTK_WIDGET_STATE (GTK_WIDGET (drawing));
+	GtkStateType state;
+	if (drawing == NULL || drawing->surface == NULL)
+		return FALSE;
 	drawing->renderContext->cr =
-	    gdk_cairo_create (GDK_DRAWABLE (drawing->pixmap));
+	    gdk_cairo_create (GDK_DRAWABLE (drawing->surface));
+
+	state = gtk_widget_get_state (GTK_WIDGET (drawing));
 	drawing->renderContext->dark_color =
-	    &GTK_WIDGET (drawing)->style->dark[state];
+	    &gtk_widget_get_style (GTK_WIDGET (drawing))->dark[state];
+	return TRUE;
 }
 
 static void
@@ -1314,26 +1324,28 @@ destroy_cairo (GkbdKeyboardDrawing * drawing)
 static void
 draw_keyboard (GkbdKeyboardDrawing * drawing)
 {
-	GtkStateType state = GTK_WIDGET_STATE (GTK_WIDGET (drawing));
-	gint pixw, pixh;
+	GtkStateType state = gtk_widget_get_state (GTK_WIDGET (drawing));
+	GtkStyle *style = gtk_widget_get_style (GTK_WIDGET (drawing));
+	GtkAllocation allocation;
 
 	if (!drawing->xkb)
 		return;
 
-	pixw = GTK_WIDGET (drawing)->allocation.width;
-	pixh = GTK_WIDGET (drawing)->allocation.height;
+	gtk_widget_get_allocation (GTK_WIDGET (drawing), &allocation);
 
-	drawing->pixmap =
-	    gdk_pixmap_new (GTK_WIDGET (drawing)->window, pixw, pixh, -1);
+	drawing->surface =
+	    gdk_window_create_similar_surface (gtk_widget_get_window (GTK_WIDGET (drawing)),
+			    allocation.width, allocation.height, -1);
 
-	/* blank background */
-	gdk_draw_rectangle (drawing->pixmap,
-			    GTK_WIDGET (drawing)->style->base_gc[state],
-			    TRUE, 0, 0, pixw, pixh);
+	if (create_cairo (drawing)) {
+		/* blank background */
+		gdk_cairo_set_source_color (drawing->renderContext->cr,
+					    &style->base[state]);
+		cairo_paint (drawing->renderContext->cr);
 
-	create_cairo (drawing);
-	draw_keyboard_to_context (drawing->renderContext, drawing);
-	destroy_cairo (drawing);
+		draw_keyboard_to_context (drawing->renderContext, drawing);
+		destroy_cairo (drawing);
+	}
 }
 
 static void
@@ -1349,8 +1361,9 @@ alloc_render_context (GkbdKeyboardDrawing * drawing)
 	pango_layout_set_ellipsize (context->layout, PANGO_ELLIPSIZE_END);
 
 	context->font_desc =
-	    pango_font_description_copy (GTK_WIDGET (drawing)->
-					 style->font_desc);
+	    pango_font_description_copy (gtk_widget_get_style
+					 (GTK_WIDGET
+					  (drawing))->font_desc);
 	context->angle = 0;
 	context->scale_numerator = 1;
 	context->scale_denominator = 1;
@@ -1368,31 +1381,29 @@ free_render_context (GkbdKeyboardDrawing * drawing)
 }
 
 static gboolean
-expose_event (GtkWidget * widget,
-	      GdkEventExpose * event, GkbdKeyboardDrawing * drawing)
+draw (GtkWidget * widget,
+      cairo_t *cr, GkbdKeyboardDrawing * drawing)
 {
-	GtkStateType state = GTK_WIDGET_STATE (GTK_WIDGET (drawing));
+	GtkAllocation allocation;
 
 	if (!drawing->xkb)
 		return FALSE;
 
-	if (drawing->pixmap == NULL)
+	if (drawing->surface == NULL)
 		return FALSE;
 
-	gdk_draw_drawable (widget->window,
-			   widget->style->fg_gc[state],
-			   drawing->pixmap,
-			   event->area.x, event->area.y,
-			   event->area.x, event->area.y,
-			   event->area.width, event->area.height);
+        cairo_set_source_surface (cr, drawing->surface, 0, 0);
+        cairo_paint (cr);
 
-	if (GTK_WIDGET_HAS_FOCUS (widget))
-		gtk_paint_focus (widget->style, widget->window,
-				 GTK_WIDGET_STATE (widget), &event->area,
+	if (gtk_widget_has_focus (widget)) {
+		gtk_widget_get_allocation (widget, &allocation);
+		gtk_paint_focus (gtk_widget_get_style (widget),
+				 cr,
+				 gtk_widget_get_state (widget),
 				 widget, "keyboard-drawing",
-				 0, 0,
-				 widget->allocation.width,
-				 widget->allocation.height);
+				 0, 0, allocation.width,
+				 allocation.height);
+	}
 
 	return FALSE;
 }
@@ -1452,9 +1463,9 @@ size_allocate (GtkWidget * widget,
 {
 	GkbdKeyboardDrawingRenderContext *context = drawing->renderContext;
 
-	if (drawing->pixmap) {
-		g_object_unref (drawing->pixmap);
-		drawing->pixmap = NULL;
+	if (drawing->surface) {
+                cairo_surface_destroy (drawing->surface);
+		drawing->surface = NULL;
 	}
 
 	if (!context_setup_scaling (context, drawing,
@@ -1492,10 +1503,12 @@ key_event (GtkWidget * widget,
 
 	key->pressed = (event->type == GDK_KEY_PRESS);
 
-	create_cairo (drawing);
-	draw_key (drawing->renderContext, drawing, key);
-	redraw_overlapping_doodads (drawing->renderContext, drawing, key);
-	destroy_cairo (drawing);
+	if (create_cairo (drawing)) {
+		draw_key (drawing->renderContext, drawing, key);
+		redraw_overlapping_doodads (drawing->renderContext,
+					    drawing, key);
+		destroy_cairo (drawing);
+	}
 
 	invalidate_key_region (drawing, key);
 	return FALSE;
@@ -1520,16 +1533,18 @@ unpress_keys (GkbdKeyboardDrawing * drawing)
 	if (!drawing->xkb)
 		return FALSE;
 
-	create_cairo (drawing);
-	for (i = drawing->xkb->min_key_code;
-	     i <= drawing->xkb->max_key_code; i++)
-		if (drawing->keys[i].pressed) {
-			drawing->keys[i].pressed = FALSE;
-			draw_key (drawing->renderContext, drawing,
-				  drawing->keys + i);
-			invalidate_key_region (drawing, drawing->keys + i);
-		}
-	destroy_cairo (drawing);
+	if (create_cairo (drawing)) {
+		for (i = drawing->xkb->min_key_code;
+		     i <= drawing->xkb->max_key_code; i++)
+			if (drawing->keys[i].pressed) {
+				drawing->keys[i].pressed = FALSE;
+				draw_key (drawing->renderContext, drawing,
+					  drawing->keys + i);
+				invalidate_key_region (drawing,
+						       drawing->keys + i);
+			}
+		destroy_cairo (drawing);
+	}
 
 	return FALSE;
 }
@@ -1665,8 +1680,8 @@ init_keys_and_doodads (GkbdKeyboardDrawing * drawing)
 				    drawing->xkb->geom->shapes +
 				    xkbkey->shape_ndx;
 				guint keycode = find_keycode (drawing,
-							      xkbkey->
-							      name.name);
+							      xkbkey->name.
+							      name);
 
 				if (keycode == INVALID_KEYCODE)
 					continue;
@@ -1777,9 +1792,8 @@ init_colors (GkbdKeyboardDrawing * drawing)
 
 	for (i = 0; i < drawing->xkb->geom->num_colors; i++) {
 		result =
-		    parse_xkb_color_spec (drawing->xkb->geom->
-					  colors[i].spec,
-					  drawing->colors + i);
+		    parse_xkb_color_spec (drawing->xkb->geom->colors[i].
+					  spec, drawing->colors + i);
 
 		if (!result)
 			g_warning
@@ -1832,6 +1846,39 @@ alloc_cdik (GkbdKeyboardDrawing * drawing)
 		    drawing->xkb->max_key_code + 1);
 }
 
+static void
+process_indicators_state_notify (XkbIndicatorNotifyEvent * iev,
+				 GkbdKeyboardDrawing * drawing)
+{
+	/* Good question: should we track indicators when the keyboard is
+	   NOT really taken from the screen */
+	gint i;
+
+	for (i = 0; i <= drawing->xkb->indicators->phys_indicators; i++)
+		if (drawing->physical_indicators[i] != NULL
+		    && (iev->changed & 1 << i)) {
+			gint state = (iev->state & 1 << i) != FALSE;
+
+			if ((state && !drawing->physical_indicators[i]->on)
+			    || (!state
+				&& drawing->physical_indicators[i]->on)) {
+				drawing->physical_indicators[i]->on =
+				    state;
+				if (create_cairo (drawing)) {
+					draw_doodad
+					    (drawing->renderContext,
+					     drawing,
+					     drawing->physical_indicators
+					     [i]);
+					destroy_cairo (drawing);
+				}
+				invalidate_indicator_doodad_region
+				    (drawing,
+				     drawing->physical_indicators[i]);
+			}
+		}
+}
+
 static GdkFilterReturn
 xkb_state_notify_event_filter (GdkXEvent * gdkxev,
 			       GdkEvent * event,
@@ -1845,6 +1892,7 @@ xkb_state_notify_event_filter (GdkXEvent * gdkxev,
 
 	if (((XEvent *) gdkxev)->type == drawing->xkb_event_type) {
 		XkbEvent *kev = (XkbEvent *) gdkxev;
+		GtkAllocation allocation;
 		switch (kev->any.xkb_type) {
 		case XkbStateNotify:
 			if (((kev->state.changed & modifier_change_mask) &&
@@ -1858,10 +1906,12 @@ xkb_state_notify_event_filter (GdkXEvent * gdkxev,
 				    g_new0 (GkbdKeyboardDrawingKey,
 					    drawing->xkb->max_key_code +
 					    1);
+
+				gtk_widget_get_allocation (GTK_WIDGET
+							   (drawing),
+							   &allocation);
 				size_allocate (GTK_WIDGET (drawing),
-					       &(GTK_WIDGET
-						 (drawing)->allocation),
-					       drawing);
+					       &allocation, drawing);
 
 				init_keys_and_doodads (drawing);
 				init_colors (drawing);
@@ -1870,49 +1920,11 @@ xkb_state_notify_event_filter (GdkXEvent * gdkxev,
 
 		case XkbIndicatorStateNotify:
 			{
-				/* Good question: should we track indicators when the keyboard is
-				   NOT really taken from the screen */
-				XkbIndicatorNotifyEvent *iev =
-				    &((XkbEvent *) gdkxev)->indicators;
-				gint i;
-
-				for (i = 0;
-				     i <=
-				     drawing->xkb->
-				     indicators->phys_indicators; i++)
-					if (drawing->physical_indicators[i]
-					    != NULL
-					    && (iev->changed & 1 << i)) {
-						gint state =
-						    (iev->state & 1 << i)
-						    != FALSE;
-
-						if ((state
-						     &&
-						     !drawing->physical_indicators
-						     [i]->on) || (!state
-								  &&
-								  drawing->physical_indicators
-								  [i]->on))
-						{
-							drawing->physical_indicators
-							    [i]->on =
-							    state;
-							create_cairo
-							    (drawing);
-							draw_doodad
-							    (drawing->renderContext,
-							     drawing,
-							     drawing->physical_indicators
-							     [i]);
-							destroy_cairo
-							    (drawing);
-							invalidate_indicator_doodad_region
-							    (drawing,
-							     drawing->physical_indicators
-							     [i]);
-						}
-					}
+				process_indicators_state_notify (&
+								 ((XkbEvent
+								   *)
+								  gdkxev)->
+indicators, drawing);
 			}
 			break;
 
@@ -1954,7 +1966,9 @@ destroy (GkbdKeyboardDrawing * drawing)
 		drawing->idle_redraw = 0;
 	}
 
-	g_object_unref (drawing->pixmap);
+        if (drawing->surface != NULL) {
+                cairo_surface_destroy (drawing->surface);
+        }
 }
 
 static void
@@ -1991,7 +2005,7 @@ gkbd_keyboard_drawing_init (GkbdKeyboardDrawing * drawing)
 		drawing->screen_num =
 		    gdk_screen_get_number (gdk_screen_get_default ());
 
-	drawing->pixmap = NULL;
+	drawing->surface = NULL;
 	alloc_render_context (drawing);
 
 	drawing->keyboard_items = NULL;
@@ -2018,7 +2032,7 @@ gkbd_keyboard_drawing_init (GkbdKeyboardDrawing * drawing)
 
 	XkbGetNames (drawing->display, XkbAllNamesMask, drawing->xkb);
 	drawing->l3mod = XkbKeysymToModifiers (drawing->display,
-					       GDK_ISO_Level3_Shift);
+					       GDK_KEY_ISO_Level3_Shift);
 
 	drawing->xkbOnDisplay = TRUE;
 
@@ -2046,14 +2060,14 @@ gkbd_keyboard_drawing_init (GkbdKeyboardDrawing * drawing)
 	init_colors (drawing);
 
 	/* required to get key events */
-	GTK_WIDGET_SET_FLAGS (GTK_WIDGET (drawing), GTK_CAN_FOCUS);
+	gtk_widget_set_can_focus (GTK_WIDGET (drawing), TRUE);
 
 	gtk_widget_set_events (GTK_WIDGET (drawing),
 			       GDK_EXPOSURE_MASK | GDK_KEY_PRESS_MASK |
 			       GDK_KEY_RELEASE_MASK | GDK_BUTTON_PRESS_MASK
 			       | GDK_FOCUS_CHANGE_MASK);
-	g_signal_connect (G_OBJECT (drawing), "expose-event",
-			  G_CALLBACK (expose_event), drawing);
+	g_signal_connect (G_OBJECT (drawing), "draw",
+			  G_CALLBACK (draw), drawing);
 	g_signal_connect_after (G_OBJECT (drawing), "key-press-event",
 				G_CALLBACK (key_event), drawing);
 	g_signal_connect_after (G_OBJECT (drawing), "key-release-event",
@@ -2137,24 +2151,6 @@ gkbd_keyboard_drawing_set_mods (GkbdKeyboardDrawing * drawing, guint mods)
 	}
 }
 
-/* returns a pixbuf with the keyboard drawing at the current pixel size
- * (which can then be saved to disk, etc) */
-GdkPixbuf *
-gkbd_keyboard_drawing_get_pixbuf (GkbdKeyboardDrawing * drawing)
-{
-	GkbdKeyboardDrawingRenderContext *context = drawing->renderContext;
-
-	if (drawing->pixmap == NULL)
-		draw_keyboard (drawing);
-
-	return gdk_pixbuf_get_from_drawable (NULL, drawing->pixmap, NULL,
-					     0, 0, 0, 0,
-					     xkb_to_pixmap_coord (context,
-								  drawing->xkb->geom->width_mm),
-					     xkb_to_pixmap_coord (context,
-								  drawing->xkb->geom->height_mm));
-}
-
 /**
  * gkbd_keyboard_drawing_render:
  * @drawing: keyboard layout to render
@@ -2183,11 +2179,12 @@ gkbd_keyboard_drawing_render (GkbdKeyboardDrawing * drawing,
 		cr,
 		drawing->renderContext->angle,
 		layout,
-		pango_font_description_copy (GTK_WIDGET (drawing)->
-					     style->font_desc),
+		pango_font_description_copy (gtk_widget_get_style
+					     (GTK_WIDGET
+					      (drawing))->font_desc),
 		1, 1,
-		&GTK_WIDGET (drawing)->style->dark[GTK_WIDGET_STATE
-						   (GTK_WIDGET (drawing))]
+		&gtk_widget_get_style (GTK_WIDGET (drawing))->dark
+		    [gtk_widget_get_state (GTK_WIDGET (drawing))]
 	};
 
 	if (!context_setup_scaling (&context, drawing, width, height,
@@ -2206,6 +2203,8 @@ gboolean
 gkbd_keyboard_drawing_set_keyboard (GkbdKeyboardDrawing * drawing,
 				    XkbComponentNamesRec * names)
 {
+	GtkAllocation allocation;
+
 	free_cdik (drawing);
 	if (drawing->xkb)
 		XkbFreeKeyboard (drawing->xkb, 0, TRUE);	/* free_all = TRUE */
@@ -2242,8 +2241,8 @@ gkbd_keyboard_drawing_set_keyboard (GkbdKeyboardDrawing * drawing,
 	init_keys_and_doodads (drawing);
 	init_colors (drawing);
 
-	size_allocate (GTK_WIDGET (drawing),
-		       &(GTK_WIDGET (drawing)->allocation), drawing);
+	gtk_widget_get_allocation (GTK_WIDGET (drawing), &allocation);
+	size_allocate (GTK_WIDGET (drawing), &allocation, drawing);
 	gtk_widget_queue_draw (GTK_WIDGET (drawing));
 
 	return TRUE;
@@ -2385,7 +2384,7 @@ gkbd_keyboard_drawing_draw_page (GtkPrintOperation * operation,
 	gdouble dpi_y = gtk_print_context_get_dpi_y (context);
 	gchar *header;
 
-	gtk_print_operation_set_unit (operation, GTK_PIXELS);
+	gtk_print_operation_set_unit (operation, GTK_UNIT_PIXEL);
 
 	header = g_strdup_printf
 	    (_("Keyboard layout \"%s\"\n"
@@ -2443,3 +2442,163 @@ gkbd_keyboard_drawing_print (GkbdKeyboardDrawing * drawing,
 
 	g_object_unref (print);
 }
+
+static void
+show_layout_response (GtkWidget * dialog, gint resp)
+{
+	GdkRectangle rect;
+	GtkWidget *kbdraw;
+	const gchar *groupName;
+
+	switch (resp) {
+	case GTK_RESPONSE_HELP:
+		gtk_show_uri (gtk_widget_get_screen (GTK_WIDGET (dialog)),
+			      "ghelp:gswitchit?layout-view",
+			      gtk_get_current_event_time (), NULL);
+		return;
+	case GTK_RESPONSE_CLOSE:
+		gtk_window_get_position (GTK_WINDOW (dialog), &rect.x,
+					 &rect.y);
+		gtk_window_get_size (GTK_WINDOW (dialog), &rect.width,
+				     &rect.height);
+		gkbd_preview_save_position (&rect);
+		gtk_widget_destroy (dialog);
+		break;
+	case GTK_RESPONSE_PRINT:
+		kbdraw =
+		    GTK_WIDGET (g_object_get_data
+				(G_OBJECT (dialog), "kbdraw"));
+		groupName =
+		    (const gchar *) g_object_get_data (G_OBJECT (dialog),
+						       "groupName");
+		gkbd_keyboard_drawing_print (GKBD_KEYBOARD_DRAWING
+					     (kbdraw), GTK_WINDOW (dialog),
+					     groupName ? groupName :
+					     _("Unknown"));
+	}
+}
+
+GtkWidget *
+gkbd_keyboard_drawing_new_dialog (gint group, gchar * group_name)
+{
+	static GkbdKeyboardDrawingGroupLevel groupsLevels[] = { {
+								 0, 1}, {
+									 0,
+									 3},
+	{
+	 0, 0}, {
+		 0, 2}
+	};
+	static GkbdKeyboardDrawingGroupLevel *pGroupsLevels[] = {
+		groupsLevels, groupsLevels + 1, groupsLevels + 2,
+		groupsLevels + 3
+	};
+
+	GtkBuilder *builder;
+	GtkWidget *dialog, *kbdraw;
+	XkbComponentNamesRec component_names;
+	XklConfigRec *xkl_data;
+	GdkRectangle *rect;
+	GError *error = NULL;
+	char title[128] = "";
+	XklEngine *engine =
+	    xkl_engine_get_instance (GDK_DISPLAY_XDISPLAY
+				     (gdk_display_get_default ()));
+
+	builder = gtk_builder_new ();
+	gtk_builder_add_from_file (builder, UIDIR "/show-layout.ui",
+				   &error);
+
+	if (error) {
+		g_error ("building ui from %s failed: %s",
+			 UIDIR "/show-layout.ui", error->message);
+		g_clear_error (&error);
+	}
+
+
+	dialog =
+	    GTK_WIDGET (gtk_builder_get_object
+			(builder, "gswitchit_layout_view"));
+	kbdraw = gkbd_keyboard_drawing_new ();
+
+	snprintf (title, sizeof (title), _("Keyboard Layout \"%s\""),
+		  group_name);
+	gtk_window_set_title (GTK_WINDOW (dialog), title);
+	g_object_set_data_full (G_OBJECT (dialog), "group_name",
+				g_strdup (group_name), g_free);
+
+	gkbd_keyboard_drawing_set_groups_levels (GKBD_KEYBOARD_DRAWING
+						 (kbdraw), pGroupsLevels);
+
+	xkl_data = xkl_config_rec_new ();
+	if (xkl_config_rec_get_from_server (xkl_data, engine)) {
+		int num_layouts = g_strv_length (xkl_data->layouts);
+		int num_variants = g_strv_length (xkl_data->variants);
+		if (group >= 0 && group < num_layouts
+		    && group < num_variants) {
+			char *l = g_strdup (xkl_data->layouts[group]);
+			char *v = g_strdup (xkl_data->variants[group]);
+			char **p;
+			int i;
+
+			if ((p = xkl_data->layouts) != NULL)
+				for (i = num_layouts; --i >= 0;)
+					g_free (*p++);
+
+			if ((p = xkl_data->variants) != NULL)
+				for (i = num_variants; --i >= 0;)
+					g_free (*p++);
+
+			xkl_data->layouts =
+			    g_realloc (xkl_data->layouts,
+				       sizeof (char *) * 2);
+			xkl_data->variants =
+			    g_realloc (xkl_data->variants,
+				       sizeof (char *) * 2);
+			xkl_data->layouts[0] = l;
+			xkl_data->variants[0] = v;
+			xkl_data->layouts[1] = xkl_data->variants[1] =
+			    NULL;
+		}
+
+		if (xkl_xkb_config_native_prepare
+		    (engine, xkl_data, &component_names)) {
+			gkbd_keyboard_drawing_set_keyboard
+			    (GKBD_KEYBOARD_DRAWING (kbdraw),
+			     &component_names);
+			xkl_xkb_config_native_cleanup (engine,
+						       &component_names);
+		}
+	}
+	g_object_unref (G_OBJECT (xkl_data));
+
+	g_object_set_data (G_OBJECT (dialog), "builderData", builder);
+	g_signal_connect (G_OBJECT (dialog), "response",
+			  G_CALLBACK (show_layout_response), NULL);
+
+	rect = gkbd_preview_load_position ();
+	if (rect != NULL) {
+		gtk_window_move (GTK_WINDOW (dialog), rect->x, rect->y);
+		gtk_window_resize (GTK_WINDOW (dialog), rect->width,
+				   rect->height);
+		g_free (rect);
+	} else
+		gtk_window_resize (GTK_WINDOW (dialog), 700, 400);
+
+	gtk_window_set_resizable (GTK_WINDOW (dialog), TRUE);
+
+	gtk_container_add (GTK_CONTAINER
+			   (gtk_builder_get_object
+			    (builder, "preview_vbox")), kbdraw);
+
+	g_object_set_data (G_OBJECT (dialog), "kbdraw", kbdraw);
+
+	g_signal_connect_swapped (dialog, "destroy",
+				  G_CALLBACK (g_object_unref),
+				  g_object_get_data (G_OBJECT (dialog),
+						     "builderData"));
+
+	gtk_widget_show_all (dialog);
+
+	return dialog;
+}
